import shutil
import zipfile
from pathlib import Path


def zip_dir(dir_path, archive_path: str):
    """
    Zips contents of a directory without the root directory.
    Example usage:
    zip_dir('dummy_dir', 'code.zip')
    will archive contents of directory 'dummy_dir' into a file 'code.zip' in CWD
    :param dir_path: Path to directory to compress.
    :param archive_path: Path to the zip file.
    """
    archive_path = str(archive_path)
    if archive_path.endswith('.zip'):
        archive_path = archive_path[:-4]  # remove .zip
    else:
        raise RuntimeError(f"archive_path {archive_path} must end with '.zip'!")

    shutil.make_archive(archive_path, 'zip', root_dir=dir_path, base_dir='.')


def add_file_to_zip(archive_path, file_path):
    with zipfile.ZipFile(archive_path, "a") as zf:
        zf.write(file_path, Path(file_path).name)



import logging
import os
from pathlib import Path

import boto3

# insert Cortex Serving Client bucket name
BUCKET_NAME = os.environ["CSC_BUCKET_NAME"]
BUCKET_SSE_KEY = os.environ['CSC_S3_SSE_KEY']

logger = logging.getLogger(__name__)


def _get_extra_args(sse_key):
    return {"SSECustomerAlgorithm": "AES256", "SSECustomerKey": sse_key}


def upload_file(local_filepath, s3_path, bucket_name=BUCKET_NAME, sse_key=BUCKET_SSE_KEY, verbose=True):
    local_filepath = str(local_filepath)
    s3_path = str(s3_path)

    if verbose:
        logger.info(f"Uploading local file {local_filepath} -> {bucket_name}:{s3_path}")

    session = boto3.session.Session()
    session.client("s3").upload_file(
        local_filepath, bucket_name, s3_path, ExtraArgs=_get_extra_args(sse_key)
    )


def upload_fileobj(fp, s3_path, bucket_name=BUCKET_NAME, sse_key=BUCKET_SSE_KEY, verbose=True):
    s3_path = str(s3_path)

    if verbose:
        logger.info(f"Uploading local file object -> {bucket_name}:{s3_path}")

    session = boto3.session.Session()
    session.client("s3").upload_fileobj(fp, bucket_name, s3_path, ExtraArgs=_get_extra_args(sse_key))


def download_fileobj(s3_path, fp, bucket_name=BUCKET_NAME, sse_key=BUCKET_SSE_KEY, verbose=True):
    s3_path = str(s3_path)

    if verbose:
        logger.info(f"Downloading file object from {bucket_name}:{s3_path}")

    session = boto3.session.Session()
    session.client("s3").download_fileobj(bucket_name, s3_path, fp, ExtraArgs=_get_extra_args(sse_key))


def download_file(s3_path, local_filepath, bucket_name=BUCKET_NAME, sse_key=BUCKET_SSE_KEY, overwrite=True):
    local_filepath = str(local_filepath)
    s3_path = str(s3_path)

    if Path(local_filepath).exists() and not overwrite:
        logger.info(f"{local_filepath} exists and overwrite=False, skipping")
    else:
        Path(local_filepath).parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"Downloading file from {bucket_name}:{s3_path} -> {local_filepath}")

        session = boto3.session.Session()
        session.client("s3").download_file(
            bucket_name, s3_path, local_filepath, ExtraArgs=_get_extra_args(sse_key)
        )



import importlib
import json
import logging
import os
import re

from fastapi import FastAPI
from starlette import status
from starlette.requests import Request
from starlette.responses import PlainTextResponse, JSONResponse


JOB_COMPLETE_PAYLOAD = '"job_complete"'
ON_JOB_COMPLETE_PATH = "/on-job-complete"


def get_class(module_path: str, class_name: str):
    """
    @param module_path: cls.__module__ = 'predictors.my_predictor'
    @param class_name: cls.__name__ = 'MyPythonPredictor'
    @return: The imported class ready to be instantiated.
    """
    module = importlib.import_module(module_path)
    return getattr(module, class_name)


app = FastAPI()
app.ready = False
app.predictor = None
logger = logging.getLogger(__name__)


@app.on_event("startup")
def startup():
    with open("predictor_config.json", "r") as f:
        config = json.load(f)

    predictor_file_path = os.environ['CSC_PREDICTOR_PATH']
    predictor_cls_name = os.environ['CSC_PREDICTOR_CLASS_NAME']
    predictor_cls = get_class(predictor_file_path, predictor_cls_name)
    app.predictor = predictor_cls(config)
    app.ready = True


@app.get("/healthz")
def healthz():
    if app.ready:
        return PlainTextResponse("ok")
    return PlainTextResponse("service unavailable", status_code=status.HTTP_503_SERVICE_UNAVAILABLE)


@app.post("/")
def handle_post_or_batch(request: Request):
    response = app.predictor.predict(request.state.payload)
    if response is not None:
        return response


@app.post(ON_JOB_COMPLETE_PATH)
def on_job_complete():
    app.predictor.on_job_complete()


@app.middleware("http")
async def parse_payload(request: Request, call_next):
    content_type = request.headers.get("content-type", "").lower()

    if request.url.path.endswith(ON_JOB_COMPLETE_PATH) or content_type.startswith("text/plain"):
        try:
            body = await get_text_body(content_type, request)
            request.state.payload = body

        except Exception as e:
            log_exception(content_type, request)
            return PlainTextResponse(content=str(e), status_code=400)

    elif content_type.startswith("multipart/form") or content_type.startswith(
        "application/x-www-form-urlencoded"
    ):
        try:
            request.state.payload = await request.form()

        except Exception as e:
            log_exception(content_type, request)
            return PlainTextResponse(content=str(e), status_code=400)

    elif content_type.startswith("application/json"):
        try:
            request.state.payload = await request.json()

        except json.JSONDecodeError as e:
            log_exception(content_type, request)
            return JSONResponse(content={"error": str(e)}, status_code=400)

        except Exception as e:
            log_exception(content_type, request)
            return JSONResponse(content={"error": str(e)}, status_code=400)

    else:
        request.state.payload = await request.body()

    response = await call_next(request)
    return response


async def get_text_body(content_type, request: Request) -> str:
    charset = "utf-8"
    matches = re.findall(r"charset=(\S+)", content_type)
    if len(matches) > 0:
        charset = matches[-1].rstrip(";")

    body = (await request.body()).decode(charset)
    return body


def log_exception(content_type, request):
    logger.exception(f"Couldn't parse {request.method} request of content type {content_type} to {request.url}")
